@@ -31,6 +31,7 @@
 
 #include "pam.h"
 #include "shhopt.h"
+#include "mallocvar.h"
 
 enum MediaSize {
     eLetterPaper,
@@ -215,7 +216,7 @@ parseCommandLine ( int argc, char ** argv,
         cmdlineP->format = eLetterPaper;
 
     if (argc-1 < 1) {
-        cmdlineP->sourceP = (InputSource *)malloc(sizeof(InputSource));
+        MALLOCVAR(cmdlineP->sourceP);
         cmdlineP->sourceP->name = "-";
     } else {
         int i;
@@ -224,7 +225,7 @@ parseCommandLine ( int argc, char ** argv,
         nextLinkP = &cmdlineP->sourceP;
         for (i = 1; i < argc; ++i) {
             InputSource * sourceP;
-            sourceP = (InputSource *)malloc(sizeof(InputSource));
+            MALLOCVAR(sourceP);
             sourceP->name = argv[i];
             *nextLinkP = sourceP;
             nextLinkP = &sourceP->next;
@@ -264,8 +265,10 @@ XY_RLEreset(XY_rle *rle)
 }
 static XY_rle * 
 XY_RLEnew(int size) {
-    XY_rle *rle=(XY_rle*)malloc(sizeof(*rle));
-    if(rle==0)
+    XY_rle *rle;
+
+    MALLOCVAR(rle);
+    if(rle==NULL)
         return rle;
     if(size<1024)
         size=1024;
@@ -1248,7 +1251,7 @@ print_pages(int                 const s,
                 ++pageNum;
                 pm_message("Processing File %u, Page %u", sourceNum, pageNum);
 
-                pnm_readpaminit(in_file, &pam, sizeof(pam));
+                pnm_readpaminit(in_file, &pam, PAM_STRUCT_SIZE(tuple_type));
                 
                 pnm_to_pclimg(&pam, &pcl_img, colorok);
 
