@@ -29,14 +29,13 @@
 #include <assert.h>
 
 #include "pam.h"
-#include "pm_gamma.h"
 #include "shhopt.h"
 #include "mallocvar.h"
 
 enum sampleScale {INTENSITY_SAMPLE, GAMMA_SAMPLE};
 /* This indicates a scale for a PAM sample value.  INTENSITY_SAMPLE means
    the value is proportional to light intensity; GAMMA_SAMPLE means the 
-   value is gamma-corrected as defined in the PGM/PPM spec.  In both
+   value is gamma-adjusted as defined in the PGM/PPM spec.  In both
    scales, the values are continuous and normalized to the range 0..1.
 
    This scale has no meaning if the PAM is not a visual image.  
@@ -380,11 +379,12 @@ enum scaleType {SCALE_SEPARATE, SCALE_BOXFIT, SCALE_BOXFILL, SCALE_PIXELMAX};
 
        SCALE_BOXFIT means specify height and width of a box, and the
        image must be scaled, preserving aspect ratio, to the largest
-       size that will fit in the box.
+       size that will fit in the box.  Some of the box may be empty.
 
        SCALE_BOXFILL means specify height and width of a box, and the
-       image must be scale, preserving aspect ratio, to the smallest
-       size that completely fills the box.  
+       image must be scaled, preserving aspect ratio, to the smallest
+       size that completely fills the box.  Some of the image may be
+       outside the box.
 
        SCALE_PIXELMAX means specify the maximum number of pixels the result
        should have and scale preserving aspect ratio and maximizing image
@@ -574,15 +574,20 @@ parseFilespecOnlyParms(int                   const argc,
 
 
 static void 
-parse_command_line(int argc, 
-                   char ** argv, 
-                   struct cmdlineInfo  * const cmdline_p) {
+parseCommandLine(int argc, 
+                 char ** argv, 
+                 struct cmdlineInfo  * const cmdlineP) {
 /* --------------------------------------------------------------------------
-   Note that the file spec array we return is stored in the storage that
-   was passed to us as the argv array.
+   Parse program command line described in Unix standard form by argc
+   and argv.  Return the information in the options as *cmdlineP.  
+
+   If command line is internally inconsistent (invalid options, etc.),
+   issue error message to stderr and abort program.
+
+   Note that the strings we return are stored in the storage that
+   was passed to us as the argv array.  We also trash *argv.
 --------------------------------------------------------------------------*/
-  
-    optEntry *option_def = malloc( 100*sizeof( optEntry ) );
+    optEntry *option_def;
     /* Instructions to optParseOptions3 on how to parse our options. */
     optStruct3 opt;
   
@@ -594,7 +599,9 @@ parse_command_line(int argc,
     const char *filterOpt, *window;
     unsigned int filterSpec, windowSpec;
 
-    option_def_index = 0;   /* incremented by OPTENTRY */
+    MALLOCARRAY_NOFAIL(option_def, 100);
+
+    option_def_index = 0;   /* incremented by OPTENT3 */
     OPTENT3(0, "xsize",     OPT_UINT,    &xsize,     NULL,                 0);
     OPTENT3(0, "width",     OPT_UINT,    &xsize,     NULL,                 0);
     OPTENT3(0, "ysize",     OPT_UINT,    &ysize,     NULL,                 0);
@@ -606,11 +613,11 @@ parse_command_line(int argc,
     OPTENT3(0, "xysize",    OPT_FLAG,    NULL,       &xyfit,               0);
     OPTENT3(0, "xyfit",     OPT_FLAG,    NULL,       &xyfit,               0);
     OPTENT3(0, "xyfill",    OPT_FLAG,    NULL,       &xyfill,              0);
-    OPTENT3(0, "verbose",   OPT_FLAG,    NULL,       &cmdline_p->verbose,  0);
+    OPTENT3(0, "verbose",   OPT_FLAG,    NULL,       &cmdlineP->verbose,  0);
     OPTENT3(0, "filter",    OPT_STRING,  &filterOpt, &filterSpec,          0);
     OPTENT3(0, "window",    OPT_STRING,  &window,    &windowSpec,          0);
-    OPTENT3(0, "nomix",     OPT_FLAG,    NULL,       &cmdline_p->nomix,    0);
-    OPTENT3(0, "linear",    OPT_FLAG,    NULL,       &cmdline_p->linear,   0);
+    OPTENT3(0, "nomix",     OPT_FLAG,    NULL,       &cmdlineP->nomix,    0);
+    OPTENT3(0, "linear",    OPT_FLAG,    NULL,       &cmdlineP->linear,   0);
   
     /* Set the defaults. -1 = unspecified */
 
@@ -630,13 +637,13 @@ parse_command_line(int argc,
     opt.allowNegNum = FALSE;   /* We have no parms that are negative numbers */
 
     optParseOptions3( &argc, argv, opt, sizeof(opt), 0 );
-    /* Uses and sets argc, argv, and some of *cmdline_p and others. */
+    /* Uses and sets argc, argv, and some of *cmdlineP and others. */
 
-    if (cmdline_p->nomix && filterSpec) 
+    if (cmdlineP->nomix && filterSpec) 
         pm_error("You cannot specify both -nomix and -filter.");
 
     processFilterOptions(filterSpec, filterOpt, windowSpec, window,
-                         cmdline_p);
+                         cmdlineP);
 
     if (xsize == 0)
         pm_error("-xsize/width must be greater than zero.");
@@ -675,42 +682,42 @@ parse_command_line(int argc,
     /* Get the program parameters */
 
     if (xyfit || xyfill) {
-        cmdline_p->scaleType = xyfit ? SCALE_BOXFIT : SCALE_BOXFILL;
-        parseXyParms(argc, argv, cmdline_p);
+        cmdlineP->scaleType = xyfit ? SCALE_BOXFIT : SCALE_BOXFILL;
+        parseXyParms(argc, argv, cmdlineP);
     } else if (reduce != -1) {
-        cmdline_p->scaleType = SCALE_SEPARATE;
-        parseFilespecOnlyParms(argc, argv, cmdline_p);
-        cmdline_p->xscale = cmdline_p->yscale = 
+        cmdlineP->scaleType = SCALE_SEPARATE;
+        parseFilespecOnlyParms(argc, argv, cmdlineP);
+        cmdlineP->xscale = cmdlineP->yscale = 
             ((double) 1.0) / ((double) reduce);
         pm_message("reducing by %d gives scale factor of %f.", 
-                   reduce, cmdline_p->xscale);
+                   reduce, cmdlineP->xscale);
     } else if (pixels != -1) {
-        cmdline_p->scaleType = SCALE_PIXELMAX;
-        parseFilespecOnlyParms(argc, argv, cmdline_p);
-        cmdline_p->pixels = pixels;
+        cmdlineP->scaleType = SCALE_PIXELMAX;
+        parseFilespecOnlyParms(argc, argv, cmdlineP);
+        cmdlineP->pixels = pixels;
     } else if (xsize == -1 && xscale == -1 && ysize == -1 && yscale == -1
                && pixels == -1 && reduce == -1) {
-        cmdline_p->scaleType = SCALE_SEPARATE;
-        parseScaleParms(argc, argv, cmdline_p);
-        cmdline_p->xsize = cmdline_p->ysize = 0;
+        cmdlineP->scaleType = SCALE_SEPARATE;
+        parseScaleParms(argc, argv, cmdlineP);
+        cmdlineP->xsize = cmdlineP->ysize = 0;
     } else {
-        cmdline_p->scaleType = SCALE_SEPARATE;
-        parseFilespecOnlyParms(argc, argv, cmdline_p);
-        cmdline_p->xsize = xsize == -1 ? 0 : xsize;
-        cmdline_p->ysize = ysize == -1 ? 0 : ysize;
-        cmdline_p->xscale = xscale == -1.0 ? 0.0 : xscale;
-        cmdline_p->yscale = yscale == -1.0 ? 0.0 : yscale;
+        cmdlineP->scaleType = SCALE_SEPARATE;
+        parseFilespecOnlyParms(argc, argv, cmdlineP);
+        cmdlineP->xsize = xsize == -1 ? 0 : xsize;
+        cmdlineP->ysize = ysize == -1 ? 0 : ysize;
+        cmdlineP->xscale = xscale == -1.0 ? 0.0 : xscale;
+        cmdlineP->yscale = yscale == -1.0 ? 0.0 : yscale;
     }
 }
 
 
 
 static void 
-compute_output_dimensions(struct cmdlineInfo  const cmdline, 
-                          int                 const rows, 
-                          int                 const cols, 
-                          int *               const newrowsP, 
-                          int *               const newcolsP) {
+computeOutputDimensions(struct cmdlineInfo  const cmdline, 
+                        int                 const rows, 
+                        int                 const cols, 
+                        int *               const newrowsP, 
+                        int *               const newcolsP) {
 
     switch(cmdline.scaleType) {
     case SCALE_PIXELMAX: {
@@ -834,10 +841,10 @@ appendWeight (WLIST *WList, int index, double weight)
 
 
 static sample
-normalizedToSample(double const normalized,
-                   sample const maxval) {
+floatToSample(double const value,
+              sample const maxval) {
 
-    return MAX(0, MIN(maxval, (sample)(normalized + 0.5)));
+    return MAX(0, MIN(maxval, (sample)(value + 0.5)));
 }
 
 
@@ -1037,7 +1044,7 @@ createScanBuf(struct pam * const pamP,
     unsigned int lineNumber;
 
     scanbuf.width = pamP->width;
-    scanbuf.height = ceil(2.0*radius + 1e-7);
+    scanbuf.height = ceil(2.0*(radius+EPSILON) + 1 + EPSILON);
     MALLOCARRAY_NOFAIL(scanbuf.line, scanbuf.height);
   
     for (lineNumber = 0; lineNumber < scanbuf.height; ++lineNumber) {
@@ -1080,25 +1087,99 @@ resampleDimensionMessage(struct pam * const inpamP,
 
 
 
+static void
+addInPixel(const struct pam * const pamP,
+           tuple              const tuple,
+           float              const weight,
+           bool               const haveOpacity,
+           unsigned int       const opacityPlane,
+           double *           const accum) {
+/*----------------------------------------------------------------------------
+  Add into *accum the values from the tuple 'tuple', weighted by
+  'weight'.
+
+  Iff 'haveOpacity', Plane 'opacityPlane' of the tuple is an opacity
+  (alpha, transparency) plane.
+-----------------------------------------------------------------------------*/
+    unsigned int plane;
+
+    for (plane = 0; plane < pamP->depth; ++plane) {
+        sample adjustedForOpacity;
+        
+        if (haveOpacity && plane != opacityPlane) {
+            float const opacity = (float)tuple[opacityPlane]/pamP->maxval;
+            float const unadjusted = (float)tuple[plane]/pamP->maxval;
+
+            adjustedForOpacity = 
+                floatToSample(unadjusted * opacity, pamP->maxval);
+        } else
+            adjustedForOpacity = tuple[plane];
+        
+        accum[plane] += (double)adjustedForOpacity * weight;
+    }
+}
+
+
+
+static void
+generateOutputTuple(const struct pam * const pamP,
+                    double             const accum[],
+                    bool               const haveOpacity, 
+                    unsigned int       const opacityPlane,
+                    tuple *            const tupleP) {
+/*----------------------------------------------------------------------------
+  Convert the values accum[] accumulated for a pixel by
+  outputOneResampledRow() to a bona fide PAM tuple as *tupleP,
+  as described by *pamP.
+-----------------------------------------------------------------------------*/
+    unsigned int plane;
+
+    for (plane = 0; plane < pamP->depth; ++plane) {
+        float opacityAdjustedSample;
+
+        if (haveOpacity && plane != opacityPlane) {
+            if (accum[opacityPlane] < EPSILON) {
+                assert(accum[plane] < EPSILON);
+                opacityAdjustedSample = 0.0;
+            } else 
+                opacityAdjustedSample = accum[plane] / accum[opacityPlane];
+        } else
+            opacityAdjustedSample = accum[plane];
+
+        (*tupleP)[plane] = floatToSample(opacityAdjustedSample, pamP->maxval);
+    }
+}
+
+
+
 static void
 outputOneResampledRow(const struct pam * const outpamP,
                       SCAN               const scanbuf,
                       WLIST              const YW,
                       const WLIST *      const XWeight,
                       tuple *            const line,
-                      double *           const weight) {
+                      double *           const accum) {
 /*----------------------------------------------------------------------------
    From the data in 'scanbuf' and weights in 'YW' and 'XWeight', 
    generate one output row for the image described by *outpamP and
    output it.
 
-   'line' and 'weight' are just working space that Caller provides us
+   An output pixel is a weighted average of the pixels in a certain
+   rectangle of the input.  'YW' and 'XWeight' describe those weights
+   for each column of the row we are to output.
+
+   'line' and 'accum' are just working space that Caller provides us
    with to save us the time of allocating it.  'line' is at least big
    enough to hold an output row; 'weight' is at least outpamP->depth
    big.
 -----------------------------------------------------------------------------*/
     unsigned int col;
 
+    bool haveOpacity;           /* There is an opacity plane */
+    unsigned int opacityPlane;  /* Plane number of opacity plane, if any */
+
+    pnm_getopacity(outpamP, &haveOpacity, &opacityPlane);
+
     for (col = 0; col < outpamP->width; ++col) {
         WLIST const XW = XWeight[col];
 
@@ -1106,7 +1187,7 @@ outputOneResampledRow(const struct pam * const outpamP,
         {
             unsigned int plane;
             for (plane = 0; plane < outpamP->depth; ++plane)
-                weight[plane] = 0.0;
+                accum[plane] = 0.0;
         }
         
         for (i = 0; i < YW.nWeight; ++i) {
@@ -1118,18 +1199,15 @@ outputOneResampledRow(const struct pam * const outpamP,
             
             for (j = 0; j < XW.nWeight; ++j) {
                 int   const xp    = XW.Weight[j].index;
-                float const xw    = XW.Weight[j].weight;
                 tuple const tuple = scanbuf.line[slot].tuplerow[xp];
                 
-                unsigned int plane;
-                
-                for (plane = 0; plane < outpamP->depth; ++plane) {
-                    weight[plane] += tuple[plane] * xw * yw;
-                    line[col][plane] = 
-                        normalizedToSample(weight[plane], outpamP->maxval);
-                }                            
+                addInPixel(outpamP, tuple, yw * XW.Weight[j].weight, 
+                           haveOpacity, opacityPlane,
+                           accum);
             }
         }
+        generateOutputTuple(outpamP, accum, haveOpacity, opacityPlane, 
+                            &line[col]);
     }
     pnm_writepamrow(outpamP, line);
 }
@@ -1142,7 +1220,8 @@ resample(struct pam *     const inpamP,
          filterFunction_t const filterFunction,
          double           const filterRadius,
          filterFunction_t const windowFunction,
-         bool             const verbose) {
+         bool             const verbose,
+         bool             const linear) {
 /*---------------------------------------------------------------------------
   Resample the image in the input file, described by *inpamP,
   so as to create the image in the output file, described by *outpamP.
@@ -1164,6 +1243,10 @@ resample(struct pam *     const inpamP,
         /* This is just work space for outputOneSampleRow() */
     double * weight;
         /* This is just work space for outputOneSampleRow() */
+
+    if (linear)
+        pm_error("You cannot use the resampling scaling method on "
+                 "linear input.");
   
     createWeightListSet(inpamP->width, outpamP->width, 
                         filterFunction, filterRadius, 
@@ -1184,6 +1267,9 @@ resample(struct pam *     const inpamP,
     row_out = 0;
     for (row_in = 0; row_in < inpamP->height; ++row_in) {
         bool targetRowsReady;
+            /* There is enough information in the buffers to output at least
+               one target row.
+            */
         int slot;
     
         /* read source row */
@@ -1193,7 +1279,6 @@ resample(struct pam *     const inpamP,
 
         targetRowsReady = TRUE; /* initial assumption */
     
-        /* check if target rows are ready */
         while (row_out < outpamP->height && targetRowsReady) {
             WLIST const YW = YWeight[row_out];
 
@@ -1252,7 +1337,15 @@ accumOutputCol(struct pam * const pamP,
                tuplen       const intuplen,
                float        const fraction, 
                tuplen       const accumulator) {
+/*----------------------------------------------------------------------------
+   Add fraction 'fraction' of the pixel indicated by 'intuplen' to the
+   pixel accumulator 'accumulator'.
 
+   'intuplen' and 'accumulator' are not a standard libnetpbm tuplen.
+   It is proportional to light intensity.  The foreground color
+   component samples are proportional to light intensity, and have
+   opacity factored in.
+-----------------------------------------------------------------------------*/
     unsigned int plane;
 
     for (plane = 0; plane < pamP->depth; ++plane)
@@ -1262,12 +1355,12 @@ accumOutputCol(struct pam * const pamP,
 
 
 static void
-horizontal_scale(tuplen *     const inputtuplenrow, 
-                 tuplen *     const newtuplenrow,
-                 struct pam * const inpamP,
-                 struct pam * const outpamP,
-                 float        const xscale,
-                 float *      const stretchP) {
+horizontalScale(tuplen *     const inputtuplenrow, 
+                tuplen *     const newtuplenrow,
+                struct pam * const inpamP,
+                struct pam * const outpamP,
+                float        const xscale,
+                float *      const stretchP) {
 /*----------------------------------------------------------------------------
   Take the input row 'inputtuplenrow', decribed by *inpamP, and scale
   it by a factor of 'xscale', to create the output row 'newtuplenrow',
@@ -1302,6 +1395,7 @@ horizontal_scale(tuplen *     const inputtuplenrow,
                            newtuplenrow[newcol]);
 
             fraccolleft -= fraccoltofill;
+
             /* Set up to start filling next output column */
             ++newcol;
             fraccoltofill = 1.0;
@@ -1331,8 +1425,7 @@ horizontal_scale(tuplen *     const inputtuplenrow,
            last output column.
         */
         accumOutputCol(inpamP, inputtuplenrow[inpamP->width-1], 
-                       fraccoltofill,
-                       newtuplenrow[newcol]);
+                       fraccoltofill, newtuplenrow[newcol]);
         
         *stretchP = fraccoltofill;
     } else 
@@ -1363,7 +1456,7 @@ accumOutputRow(struct pam * const pamP,
                tuplen *     const accumulator) {
 /*----------------------------------------------------------------------------
    Take 'fraction' times the samples in row 'tuplenrow' and add it to 
-   'accumulator'.
+   'accumulator' in the same way as accumOutputCol().
 
    'fraction' is less than 1.0.
 -----------------------------------------------------------------------------*/
@@ -1371,7 +1464,6 @@ accumOutputRow(struct pam * const pamP,
 
     for (plane = 0; plane < pamP->depth; ++plane) {
         unsigned int col;
-        
         for (col = 0; col < pamP->width; ++col)
             accumulator[col][plane] += fraction * tuplenrow[col][plane];
     }
@@ -1384,24 +1476,26 @@ readARow(struct pam *     const pamP,
          tuplen *         const tuplenRow,
          enum sampleScale const scale) {
 /*----------------------------------------------------------------------------
-  Read a row from the input file described by *pamP, as values
-  proportional to light intensity.
+  Read a row from the input file described by *pamP, as values (for the
+  foreground color) proportional to light intensity, with opacity
+  included.
 
-  By contrast, libnetpbm provides only sample values that are
-  gamma-corrected, which means they are almost proportional to
-  subjective brightness, and not to light intensity.
+  By contrast, a simple libnetpbm read would give the same numbers you
+  find in a PAM: gamma-adjusted values for the foreground color
+  component and scaled as if opaque.  The latter means that full red
+  would have a red intensity of 1.0 even if the pixel is only 75%
+  opaque.  We, on the other hand, would return red intensity of .75 in
+  that case.
+
+  The opacity plane we return is the same as a simple libnetpbm read
+  would return.
 -----------------------------------------------------------------------------*/
     pnm_readpamrown(pamP, tuplenRow);
 
-    if (scale == INTENSITY_SAMPLE) {
-        unsigned int col;
-    
-        for (col = 0; col < pamP->width; ++col) {
-            unsigned int plane;
-            for (plane = 0; plane < pamP->depth; ++plane)
-                tuplenRow[col][plane] = pm_ungamma709(tuplenRow[col][plane]);
-        }
-    }
+    if (scale != INTENSITY_SAMPLE)
+        pnm_ungammarown(pamP, tuplenRow);
+
+    pnm_applyopacityrown(pamP, tuplenRow);
 }
 
 
@@ -1412,23 +1506,16 @@ writeARow(struct pam *     const pamP,
           enum sampleScale const scale) {
 /*----------------------------------------------------------------------------
   Write a row to the output file described by *pamP, from values
-  proportional to light intensity.
-
-  By contrast, libnetpbm takes only sample values that are
-  gamma-corrected, which means they are almost proportional to
-  subjective brightness, and not to light intensity.
+  proportional to light intensity with opacity included (i.e. the same
+  kind of number you would get form readARow()).
 
   We destroy *tuplenRow in the process.
 -----------------------------------------------------------------------------*/
-    if (scale == INTENSITY_SAMPLE) {
-        unsigned int col;
-    
-        for (col = 0; col < pamP->width; ++col) {
-            unsigned int plane;
-            for (plane = 0; plane < pamP->depth; ++plane)
-                tuplenRow[col][plane] = pm_gamma709(tuplenRow[col][plane]);
-        }
-    }
+    pnm_unapplyopacityrown(pamP, tuplenRow);
+
+    if (scale != INTENSITY_SAMPLE)
+        pnm_gammarown(pamP, tuplenRow);
+
     pnm_writepamrown(pamP, tuplenRow);
 }
 
@@ -1454,6 +1541,44 @@ issueStretchWarning(bool   const verbose,
 
 
 
+static void
+scaleHorizontallyAndOutputRow(struct pam * const inpamP,
+                              struct pam * const outpamP,
+                              tuplen *     const rowAccumulator,
+                              enum sampleScale const sampleScale,
+                              tuplen *     const newtuplenrow,
+                              float        const xscale,
+                              unsigned int const row,
+                              bool         const verbose) {
+/*----------------------------------------------------------------------------
+   Scale the row in 'rowAccumulator' horizontally by factor 'xscale'
+   and output it.
+
+   'newtuplenrow' is work space Caller provides us.  It is at least
+   wide enough to hold one output row.
+-----------------------------------------------------------------------------*/
+    if (outpamP->width == inpamP->width)    
+        /* shortcut X scaling */
+        writeARow(outpamP, rowAccumulator, sampleScale);
+            /* This destroys 'rowAccumulator' */
+    else {
+        float stretch;
+            
+        horizontalScale(rowAccumulator, newtuplenrow, inpamP, outpamP,
+                        xscale, &stretch);
+            
+        if (verbose && row == 0)
+            pm_message("%f of right column stretched due to "
+                       "arithmetic imprecision", 
+                       stretch);
+            
+        writeARow(outpamP, newtuplenrow, sampleScale);
+            /* This destroys 'newtuplenrow' */
+    }
+}
+
+
+
 static void
 scaleWithMixing(struct pam * const inpamP,
                 struct pam * const outpamP,
@@ -1476,6 +1601,7 @@ scaleWithMixing(struct pam * const inpamP,
   intensity, (This makes the computation a lot faster, so you might
   use this even if the samples are actually standard PAM, to get
   approximate but fast results).
+
 -----------------------------------------------------------------------------*/
     /* Here's how we think of the color mixing scaling operation:  
        
@@ -1501,6 +1627,17 @@ scaleWithMixing(struct pam * const inpamP,
 
        This works for all scale factors, both scaling up and scaling down.
        
+       For images with an opacity plane, imagine Input Pixel 0's
+       foreground is fully opaque red (1,0,0,1), and Input Pixel 1 is
+       fully transparent (foreground irrelevant) (0,0,0,0).  We make
+       Output Pixel 0's foreground fully opaque red as before.  Output
+       Pixel 1 is covered half by Input Pixel 0 and half by Input
+       Pixel 1, so it is 50% opaque; but its foreground color is still
+       red: (1,0,0,0.5).  The output foreground color is the opacity
+       and coverage weighted average of the input foreground colors,
+       and the output opacity is the coverage weighted average of the
+       input opacities.
+
        This program always stretches or squeezes the input row to be the
        same length as the output row; The output row's pixels are always
        1 unit wide.
@@ -1510,13 +1647,13 @@ scaleWithMixing(struct pam * const inpamP,
        stretch the image vertically first (same process as above, but
        in place of a single-color pixels, we have a vector of colors).
        Then we take each row this vertical stretching generates and
-       stretch it horizontally.  
-    */
+       stretch it horizontally.  */
+
     enum sampleScale const sampleScale = 
         assumeLinear ? INTENSITY_SAMPLE : GAMMA_SAMPLE;
 
     tuplen * tuplenrow;     /* An input row */
-    tuplen * newtuplenrow;  /* An output row */
+    tuplen * newtuplenrow;  /* Working space */
     float rowsleft;
     /* The number of rows of output that need to be formed from the
        current input row (the one in tuplerow[]), less the number that 
@@ -1566,12 +1703,12 @@ scaleWithMixing(struct pam * const inpamP,
                     rowsleft = yscale;
                 }
                 if (rowsleft < fracrowtofill) {
-                    accumOutputRow(inpamP, tuplenrow, rowsleft, 
+                    accumOutputRow(inpamP, tuplenrow, rowsleft,
                                    rowAccumulator);
                     fracrowtofill -= rowsleft;
                     rowsleft = 0.0;
                 } else {
-                    accumOutputRow(inpamP, tuplenrow, fracrowtofill, 
+                    accumOutputRow(inpamP, tuplenrow, fracrowtofill,
                                    rowAccumulator);
                     rowsleft = rowsleft - fracrowtofill;
                     fracrowtofill = 0.0;
@@ -1581,27 +1718,13 @@ scaleWithMixing(struct pam * const inpamP,
         }
         /* 'rowAccumulator' now contains the contents of a single
            output row, but not yet horizontally scaled.  Scale it now
-           horizontally into 'newtuplenrow' and write it out.
+           horizontally and write it out.
         */
+        scaleHorizontallyAndOutputRow(inpamP, outpamP, rowAccumulator,
+                                      sampleScale, newtuplenrow, xscale,
+                                      row, verbose);
+            /* Destroys rowAccumulator */
 
-        if (outpamP->width == inpamP->width)    
-            /* shortcut X scaling */
-            writeARow(outpamP, rowAccumulator, sampleScale);
-                /* This destroys 'rowAccumulator' */
-        else {
-            float stretch;
-            
-            horizontal_scale(rowAccumulator, newtuplenrow, inpamP, outpamP,
-                             xscale, &stretch);
-            
-            if (verbose && row == 0)
-                pm_message("%f of right column stretched due to "
-                           "arithmetic imprecision", 
-                           stretch);
-            
-            writeARow(outpamP, newtuplenrow, sampleScale);
-                /* This destroys 'newtuplenrow' */
-        }
     }
     pnm_freepamrown(rowAccumulator);
     pnm_freepamrown(newtuplenrow);
@@ -1669,11 +1792,11 @@ main(int argc, char **argv ) {
 
     pnm_init(&argc, argv);
 
-    parse_command_line(argc, argv, &cmdline);
+    parseCommandLine(argc, argv, &cmdline);
 
     ifP = pm_openr(cmdline.input_filespec);
 
-    pnm_readpaminit(ifP, &inpam, sizeof(inpam));
+    pnm_readpaminit(ifP, &inpam, PAM_STRUCT_SIZE(tuple_type));
 
     outpam = inpam;  /* initial value */
     outpam.file = stdout;
@@ -1687,8 +1810,8 @@ main(int argc, char **argv ) {
         outpam.maxval = inpam.maxval;
     }
 
-    compute_output_dimensions(cmdline, inpam.height, inpam.width,
-                              &outpam.height, &outpam.width);
+    computeOutputDimensions(cmdline, inpam.height, inpam.width,
+                            &outpam.height, &outpam.width);
 
     xscale = (float) outpam.width / inpam.width;
     yscale = (float) outpam.height / inpam.height;
@@ -1722,7 +1845,8 @@ main(int argc, char **argv ) {
             pm_message("Using general filter method");
         resample(&inpam, &outpam,
                  cmdline.filterFunction, cmdline.filterRadius,
-                 cmdline.windowFunction, cmdline.verbose);
+                 cmdline.windowFunction, cmdline.verbose,
+                 cmdline.linear);
     }
     pm_close(ifP);
     pm_close(stdout);
