@@ -16,6 +16,7 @@
 #include "pam.h"
 #include "shhopt.h"
 #include "nstring.h"
+#include "mallocvar.h"
 
 typedef struct { int f[sizeof(int) * 8 + 1]; } factorset;
 typedef struct { int x; int y; } coord;
@@ -159,8 +160,8 @@ findpack(struct pam *imgs, int n, coord *coords)
   for (cdiv = imgs[0].width, i = 1; i < n; ++i)
     cdiv = gcd(imgs[i].width, cdiv);
 
-  current = (coord *)malloc(sizeof(coord) * n);
-  set = (coord *)malloc(sizeof(coord) * n);
+  MALLOCARRAY(current, n);
+  MALLOCARRAY(set, n);
   for (i = 0; i < n; ++i)
     set[i].x = imgs[i].width,
     set[i].y = imgs[i].height;
@@ -259,9 +260,10 @@ main(int argc, char **argv)
   else
     nfiles = 1;
 
-  imgs = (struct pam*)malloc(nfiles * sizeof(struct pam));
-  coords = (coord *)malloc(nfiles * sizeof(coord));
-  names = (char **)malloc(nfiles * sizeof(char *));
+  MALLOCARRAY(imgs, nfiles);
+  MALLOCARRAY(coords, nfiles);
+  MALLOCARRAY(names, nfiles);
+  
   if (!imgs || !coords || !names)
     pm_error("out of memory");
 
@@ -287,7 +289,7 @@ main(int argc, char **argv)
     imgs[0].file = stdin;
   }
 
-  pnm_readpaminit(imgs[0].file, &imgs[0], sizeof(struct pam));
+  pnm_readpaminit(imgs[0].file, &imgs[0], PAM_STRUCT_SIZE(tuple_type));
   outimg.maxval = imgs[0].maxval;
   outimg.format = imgs[0].format;
   memcpy(outimg.tuple_type, imgs[0].tuple_type, sizeof(imgs[0].tuple_type));
@@ -295,7 +297,7 @@ main(int argc, char **argv)
 
   for (i = 1; i < nfiles; ++i)
   {
-    pnm_readpaminit(imgs[i].file, &imgs[i], sizeof(struct pam));
+    pnm_readpaminit(imgs[i].file, &imgs[i], PAM_STRUCT_SIZE(tuple_type));
     if (PAM_FORMAT_TYPE(imgs[i].format) > PAM_FORMAT_TYPE(outimg.format))
       outimg.format = imgs[i].format,
       memcpy(outimg.tuple_type, imgs[i].tuple_type, 
