@@ -5,7 +5,7 @@
    that deal with the PAM (Portable Arbitrary Format) image format.
 -----------------------------------------------------------------------------*/
 
-/* See libpbm.c for the complicated explanation of this 32/64 bit file
+/* See libpm.c for the complicated explanation of this 32/64 bit file
    offset stuff.
 */
 #define _FILE_OFFSET_BITS 64
@@ -13,6 +13,9 @@
 
 #include <string.h>
 #include <limits.h>
+#include <assert.h>
+
+#include <math.h>
 
 #include "mallocvar.h"
 #include "nstring.h"
@@ -75,7 +78,28 @@ stripeq(const char * const comparand, const char * const comparator) {
     if (*p != *q) equal = FALSE;
     p++; q++;
   }
-  return(equal);
+  return equal;
+}
+
+
+
+static unsigned int
+allocationDepth(const struct pam * const pamP) {
+
+    unsigned int retval;
+
+    if (pamP->len >= PAM_STRUCT_SIZE(allocation_depth)) {
+        if (pamP->allocation_depth == 0)
+            retval = pamP->depth;
+        else {
+            if (pamP->depth > pamP->allocation_depth)
+                pm_error("'allocationDepth' (%u) is smaller than 'depth' (%u)",
+                         pamP->allocation_depth, pamP->depth);
+            retval = pamP->allocation_depth;
+        }
+    } else
+        retval = pamP->depth;
+    return retval;
 }
 
 
@@ -96,18 +120,20 @@ validateComputableSize(struct pam * const pamP) {
    Another common operation is adding 1 or 2 to the highest row, column,
    or plane number in the image, so we make sure that's possible.
 -----------------------------------------------------------------------------*/
-    if (pamP->depth > INT_MAX/sizeof(sample))
-        pm_error("image depth (%u) too large to be processed", pamP->depth);
-    else if (pamP->depth * sizeof(sample) > INT_MAX/pamP->width)
-        pm_error("image width and depth (%u, %u) too large "
-                 "to be processed.", pamP->width, pamP->depth);
-    else if (pamP->width * (pamP->depth * sizeof(sample)) >
-             INT_MAX - pamP->depth * sizeof(tuple *))
-        pm_error("image width and depth (%u, %u) too large "
-                 "to be processed.", pamP->width, pamP->depth);
+    unsigned int const depth = allocationDepth(pamP);
 
-    if (pamP->depth > INT_MAX - 2)
-        pm_error("image depth (%u) too large to be processed", pamP->depth);
+    if (depth > INT_MAX/sizeof(sample))
+        pm_error("image depth (%u) too large to be processed", depth);
+    else if (depth * sizeof(sample) > INT_MAX/pamP->width)
+        pm_error("image width and depth (%u, %u) too large "
+                 "to be processed.", pamP->width, depth);
+    else if (pamP->width * (depth * sizeof(sample)) >
+             INT_MAX - depth * sizeof(tuple *))
+        pm_error("image width and depth (%u, %u) too large "
+                 "to be processed.", pamP->width, depth);
+    
+    if (depth > INT_MAX - 2)
+        pm_error("image depth (%u) too large to be processed", depth);
     if (pamP->width > INT_MAX - 2)
         pm_error("image width (%u) too large to be processed", pamP->width);
     if (pamP->height > INT_MAX - 2)
@@ -116,6 +142,22 @@ validateComputableSize(struct pam * const pamP) {
 
 
 
+tuple
+pnm_allocpamtuple(const struct pam * const pamP) {
+
+    tuple retval;
+
+    retval = malloc(allocationDepth(pamP) * sizeof(retval[0]));
+
+    if (retval == NULL)
+        pm_error("Out of memory allocating %u-plane tuple", 
+                 allocationDepth(pamP));
+
+    return retval;
+}
+
+
+
 int
 pnm_tupleequal(const struct pam * const pamP, 
                tuple              const comparand, 
@@ -152,13 +194,11 @@ pnm_assigntuple(const struct pam * const pamP,
 
 
 
-void
-pnm_scaletuple(const struct pam * const pamP,
-               tuple              const dest,
-               tuple              const source, 
-               sample             const newmaxval) {
-
-    /* TODO: For speed reasons, we should probably make this an inline */
+static void
+scaleTuple(const struct pam * const pamP,
+           tuple              const dest,
+           tuple              const source, 
+           sample             const newmaxval) {
 
     unsigned int plane;
     for (plane = 0; plane < pamP->depth; ++plane) 
@@ -167,18 +207,17 @@ pnm_scaletuple(const struct pam * const pamP,
 
 
 
-void 
-pnm_scaletuplerow(const struct pam * const pamP,
-                  tuple *            const destRow,
-                  tuple *            const sourceRow,
-                  sample             const newMaxval) {
+void
+pnm_scaletuple(const struct pam * const pamP,
+               tuple              const dest,
+               tuple              const source, 
+               sample             const newmaxval) {
 
-    unsigned int col;
-    for (col = 0; col < pamP->width; ++col) 
-        pnm_scaletuple(pamP, destRow[col], sourceRow[col], newMaxval);
+    scaleTuple(pamP, dest, source, newmaxval);
 }
 
- 
+
+
 void
 pnm_createBlackTuple(const struct pam * const pamP, 
                      tuple *            const blackTupleP) {
@@ -226,7 +265,7 @@ pnm_allocpamrow(const struct pam * const pamP) {
    overflow will not occur in our calculations.  NOTE: pnm_readpaminit()
    ensures this assumption is valid.
 -----------------------------------------------------------------------------*/
-    const int bytes_per_tuple = pamP->depth * sizeof(sample);
+    const int bytes_per_tuple = allocationDepth(pamP) * sizeof(sample);
     tuple * tuplerow;
 
     /* The tuple row data structure starts with 'width' pointers to
@@ -238,7 +277,7 @@ pnm_allocpamrow(const struct pam * const pamP) {
     if (tuplerow == NULL)
         pm_error("Out of memory allocating space for a tuple row of\n"
                  "%d tuples by %d samples per tuple by %d bytes per sample.",
-                 pamP->width, pamP->depth, sizeof(sample));
+                 pamP->width, allocationDepth(pamP), sizeof(sample));
 
     {
         /* Now we initialize the pointers to the individual tuples to make this
@@ -254,11 +293,70 @@ pnm_allocpamrow(const struct pam * const pamP) {
             p += bytes_per_tuple;
         }
     }
-    return(tuplerow);
+    return tuplerow;
 }
 
 
 
+static unsigned int 
+rowimagesize(const struct pam * const pamP) {
+
+    /* If repeatedly calculating this turns out to be a significant
+       performance problem, we could keep this in struct pam like
+       bytes_per_sample.
+    */
+
+    if (PAM_FORMAT_TYPE(pamP->format) == PBM_TYPE)
+        return pbm_packed_bytes(pamP->width);
+    else 
+        return (pamP->width * pamP->bytes_per_sample * pamP->depth);
+}
+
+
+
+unsigned char *
+pnm_allocrowimage(const struct pam * const pamP) {
+
+    unsigned int const rowsize = rowimagesize(pamP);
+    unsigned int const overrunSpaceNeeded = 8;
+        /* This is the number of extra bytes of space libnetpbm needs to have
+           at the end of the buffer so it can use fast, lazy algorithms.
+        */
+    unsigned int const size = rowsize + overrunSpaceNeeded;
+
+    unsigned char * retval;
+
+    retval = malloc(size);
+
+    if (retval == NULL)
+        pm_error("Unable to allocate %u bytes for a row image buffer",
+                 size);
+
+    return retval;
+}
+
+
+
+void
+pnm_freerowimage(unsigned char * const rowimage) {
+    free(rowimage);
+}
+
+
+
+void 
+pnm_scaletuplerow(const struct pam * const pamP,
+                  tuple *            const destRow,
+                  tuple *            const sourceRow,
+                  sample             const newMaxval) {
+
+    unsigned int col;
+    for (col = 0; col < pamP->width; ++col) 
+        scaleTuple(pamP, destRow[col], sourceRow[col], newMaxval);
+}
+
+ 
+
 tuple **
 pnm_allocpamarray(const struct pam * const pamP) {
     
@@ -267,17 +365,17 @@ pnm_allocpamarray(const struct pam * const pamP) {
 
     /* If the speed of this is ever an issue, it might be sped up a little
        by allocating one large chunk.
-       */
+    */
     
     MALLOCARRAY(tuplearray, pamP->height);
     if (tuplearray == NULL) 
         pm_error("Out of memory allocating the row pointer section of "
                  "a %u row array", pamP->height);
 
-    for (row = 0; row < pamP->height; row++) {
+    for (row = 0; row < pamP->height; ++row) {
         tuplearray[row] = pnm_allocpamrow(pamP);
     }
-    return(tuplearray);
+    return tuplearray;
 }
 
 
@@ -293,15 +391,27 @@ pnm_freepamarray(tuple ** const tuplearray, const struct pam * const pamP) {
 }
 
 
+
+void 
+pnm_setminallocationdepth(struct pam * const pamP,
+                          unsigned int const allocationDepth) {
+
+    pamP->allocation_depth = MAX(allocationDepth, pamP->depth);
+        
+    validateComputableSize(pamP);
+}
+
+
+
 void
-pnm_setpamrow(struct pam const pam, 
-              tuple *    const tuplerow, 
-              sample     const value) {
+pnm_setpamrow(const struct pam * const pamP,
+              tuple *            const tuplerow, 
+              sample             const value) {
 
     int col;
-    for (col = 0; col < pam.width; ++col) {
+    for (col = 0; col < pamP->width; ++col) {
         int plane;
-        for (plane = 0; plane < pam.depth; ++plane) 
+        for (plane = 0; plane < pamP->depth; ++plane) 
             tuplerow[col][plane] = value;
     }
 }
@@ -424,7 +534,7 @@ process_header_line(const char buffer[], struct pam * const pamP,
 
 
 static void
-pnm_readpaminitrest(struct pam * const pamP) {
+readpaminitrest(struct pam * const pamP) {
 /*----------------------------------------------------------------------------
    Read the rest of the PAM header (after the first line -- the magic
    number line).  Fill in all the information in *pamP.
@@ -439,11 +549,12 @@ pnm_readpaminitrest(struct pam * const pamP) {
     pamP->maxval = 0;
     pamP->tuple_type[0] = '\0';
 
-    { int c;
-    /* Read off rest of 1st line -- probably just the newline after the 
-       magic number 
-    */
-    while ((c = getc(pamP->file)) != -1 && c != '\n');
+    { 
+        int c;
+        /* Read off rest of 1st line -- probably just the newline after the 
+           magic number 
+        */
+        while ((c = getc(pamP->file)) != -1 && c != '\n');
     }    
 
     through_header = FALSE;
@@ -473,9 +584,23 @@ pnm_readpaminitrest(struct pam * const pamP) {
 
 
 
+unsigned int
+pnm_bytespersample(sample const maxval) {
+
+    if (maxval >> 8 == 0)       return 1;
+    else if (maxval >> 16 == 0) return 2;
+    else if (maxval >> 24 == 0) return 3;
+    else                        return 4;
+}
+
+
+
 void 
-pnm_readpaminit(FILE *file, struct pam * const pamP, const int size) {
-    if (size < (char*)&pamP->tuple_type - (char*)pamP) 
+pnm_readpaminit(FILE *       const file, 
+                struct pam * const pamP, 
+                int          const size) {
+
+    if (size < PAM_STRUCT_SIZE(tuple_type)) 
         pm_error("pam object passed to pnm_readpaminit() is too small.  "
                  "It must be large\n"
                  "enough to hold at least up to the "
@@ -487,6 +612,9 @@ pnm_readpaminit(FILE *file, struct pam * const pamP, const int size) {
     pamP->file = file;
     pamP->len = MIN(pamP->size, sizeof(struct pam));
 
+    if (size >= PAM_STRUCT_SIZE(allocation_depth))
+        pamP->allocation_depth = 0;
+
     /* Get magic number. */
     pamP->format = pm_readmagicnumber(file);
 
@@ -497,7 +625,7 @@ pnm_readpaminit(FILE *file, struct pam * const pamP, const int size) {
 
     switch (PAM_FORMAT_TYPE(pamP->format)) {
     case PAM_TYPE: 
-        pnm_readpaminitrest(pamP);
+        readpaminitrest(pamP);
         break;
     case PPM_TYPE: {
         pixval maxval;
@@ -528,10 +656,7 @@ pnm_readpaminit(FILE *file, struct pam * const pamP, const int size) {
         pm_error("bad magic number - not a PAM, PPM, PGM, or PBM file");
     }
     
-    if (pamP->maxval >> 8 == 0)       pamP->bytes_per_sample = 1;
-    else if (pamP->maxval >> 16 == 0) pamP->bytes_per_sample = 2;
-    else if (pamP->maxval >> 24 == 0) pamP->bytes_per_sample = 3;
-    else                              pamP->bytes_per_sample = 4;
+    pamP->bytes_per_sample = pnm_bytespersample(pamP->maxval);
 
     validateComputableSize(pamP);
 }
@@ -539,66 +664,9 @@ pnm_readpaminit(FILE *file, struct pam * const pamP, const int size) {
 
 
 void 
-pnm_readpamrow(const struct pam * const pamP, tuple* const tuplerow) {
+pnm_writepaminit(struct pam * const pamP) {
 
-    /* For speed, we don't check any of the inputs for consistency 
-       here (unless it's necessary to avoid crashing).  Any consistency
-       checking should have been done by a prior call to 
-       pnm_writepaminit().
-    */
-
-    /* Need a special case for raw PBM because it has multiple tuples (8)
-       packed into one byte.
-       */
-    if (PAM_FORMAT_TYPE(pamP->format) == PBM_TYPE) {
-        int col;
-        bit *bitrow;
-        if (pamP->depth != 1)
-            pm_error("Invalid pam structure passed to pnm_readpamrow().  "
-                     "It says PBM format, but 'depth' member is not 1.");
-        bitrow = pbm_allocrow(pamP->width);
-        pbm_readpbmrow(pamP->file, bitrow, pamP->width, pamP->format);
-        for (col = 0; col < pamP->width; col++)
-            tuplerow[col][0] = 
-                bitrow[col] == PBM_BLACK ? PAM_PBM_BLACK : PAM_PBM_WHITE;
-        pbm_freerow(bitrow);
-    } else {
-        int col;
-        for (col = 0; col < pamP->width; col++) {
-            int samp;
-            for (samp = 0; samp < pamP->depth; samp++) {
-                if (pamP->plainformat)
-                    tuplerow[col][samp] = pm_getuint(pamP->file);
-                else 
-                    tuplerow[col][samp] = 
-                        pm_getraw(pamP->file, pamP->bytes_per_sample);
-            }
-        }
-    }
-}
-
-
-
-tuple** 
-pnm_readpam(FILE *file, struct pam * const pamP, const int size) {
-
-    tuple **tuplearray;
-    int row;
-
-    pnm_readpaminit(file, pamP, size);
-    
-    tuplearray = pnm_allocpamarray(pamP);
-    
-    for (row = 0; row < pamP->height; row++) 
-        pnm_readpamrow(pamP, tuplearray[row]);
-
-    return(tuplearray);
-}
-
-
-
-void 
-pnm_writepaminit(struct pam *const pamP) {
+    const char * tupleType;
 
     if (pamP->size < pamP->len)
         pm_error("pam object passed to pnm_writepaminit() is smaller "
@@ -607,14 +675,19 @@ pnm_writepaminit(struct pam *const pamP) {
                  "(%d bytes, according to its 'len' element).",
                  pamP->size, pamP->len);
 
-    if (pamP->len < (char*)&pamP->tuple_type - (char*)pamP) 
+    if (pamP->len < PAM_STRUCT_SIZE(bytes_per_sample))
         pm_error("pam object passed to pnm_writepaminit() is too small.  "
                  "It must be large\n"
-                 "enough to hold at least up to the "
-                 "'tuple_type' member, but according\n"
-                 "to its 'len' element, it is only %d bytes long.", 
+                 "enough to hold at least up through the "
+                 "'bytes_per_sample' member, but according\n"
+                 "to its 'len' member, it is only %d bytes long.", 
                  pamP->len);
 
+    if (pamP->len < PAM_STRUCT_SIZE(tuple_type))
+        tupleType = "";
+    else
+        tupleType = pamP->tuple_type;
+
     if (pamP->maxval >> 8 == 0)       pamP->bytes_per_sample = 1;
     else if (pamP->maxval >> 16 == 0) pamP->bytes_per_sample = 2;
     else if (pamP->maxval >> 24 == 0) pamP->bytes_per_sample = 3;
@@ -627,14 +700,18 @@ pnm_writepaminit(struct pam *const pamP) {
         fprintf(pamP->file, "HEIGHT %d\n",   pamP->height);
         fprintf(pamP->file, "DEPTH %d\n",    pamP->depth);
         fprintf(pamP->file, "MAXVAL %ld\n",  pamP->maxval);
-        if (!stripeq(pamP->tuple_type, ""))
+        if (!stripeq(tupleType, ""))
             fprintf(pamP->file, "TUPLTYPE %s\n", pamP->tuple_type);
         fprintf(pamP->file, "ENDHDR\n");
         break;
-	case PPM_TYPE:
-        if (pamP->depth != 3) 
+    case PPM_TYPE:
+        /* The depth must be exact, because pnm_writepamrow() is controlled
+           by it, without regard to format.
+        */
+        if (pamP->depth != 3)
             pm_error("pnm_writepaminit() got PPM format, but depth = %d "
-                     "instead of 3, as required for PPM.", pamP->depth);
+                     "instead of 3, as required for PPM.", 
+                     pamP->depth);
         if (pamP->maxval > PPM_OVERALLMAXVAL) 
             pm_error("pnm_writepaminit() got PPM format, but maxval = %ld, "
                      "which exceeds the maximum allowed for PPM: %d", 
@@ -643,10 +720,11 @@ pnm_writepaminit(struct pam *const pamP) {
                          (pixval) pamP->maxval, 0);
         break;
 
-	case PGM_TYPE:
-        if (pamP->depth != 1) 
+    case PGM_TYPE:
+        if (pamP->depth != 1)
             pm_error("pnm_writepaminit() got PGM format, but depth = %d "
-                     "instead of 1, as required for PGM.", pamP->depth);
+                     "instead of 1, as required for PGM.",
+                     pamP->depth);
         if (pamP->maxval > PGM_OVERALLMAXVAL)
             pm_error("pnm_writepaminit() got PGM format, but maxval = %ld, "
                      "which exceeds the maximum allowed for PGM: %d", 
@@ -655,10 +733,11 @@ pnm_writepaminit(struct pam *const pamP) {
                          (gray) pamP->maxval, 0);
         break;
 
-	case PBM_TYPE:
-        if (pamP->depth != 1) 
+    case PBM_TYPE:
+        if (pamP->depth != 1)
             pm_error("pnm_writepaminit() got PBM format, but depth = %d "
-                     "instead of 1, as required for PBM.", pamP->depth);
+                     "instead of 1, as required for PBM.",
+                     pamP->depth);
         if (pamP->maxval != 1) 
             pm_error("pnm_writepaminit() got PBM format, but maxval = %ld "
                      "instead of 1, as required for PBM.", pamP->maxval);
@@ -671,61 +750,6 @@ pnm_writepaminit(struct pam *const pamP) {
     }
 }
 
-            
-
-void 
-pnm_writepamrow(const struct pam *const pamP, const tuple * const tuplerow) {
-
-    /* For speed, we don't check any of the inputs for consistency 
-       here (unless it's necessary to avoid crashing).  Any consistency
-       checking should have been done by a prior call to 
-       pnm_writepaminit().
-    */
-
-    /* Need a special case for raw PBM because it has multiple tuples (8)
-       packed into one byte.
-       */
-    if (PAM_FORMAT_TYPE(pamP->format) == PBM_TYPE) {
-        int col;
-        bit *bitrow;
-        bitrow = pbm_allocrow(pamP->width);
-        for (col = 0; col < pamP->width; col++)
-            bitrow[col] = 
-                tuplerow[col][0] == PAM_PBM_BLACK ? PBM_BLACK : PBM_WHITE;
-        pbm_writepbmrow(pamP->file, bitrow, pamP->width, 0);
-        pbm_freerow(bitrow);
-    } else {
-        int col;
-
-        for (col = 0; col < pamP->width; col++) {
-            unsigned int samp;
-            for (samp = 0; samp < pamP->depth; ++samp) {
-                /* If we allowed writing of plain format via the pam functions,
-                   we could do a putus() here to write a plain sample.  We'd
-                   also need some stuff to put in the spaces and line breaks.
-                   */
-                pm_putraw(pamP->file, 
-                          tuplerow[col][samp], 
-                          pamP->bytes_per_sample);
-            }
-        }    
-    }
-}
-
-
-
-void 
-pnm_writepam(struct pam *const pamP, tuple ** const tuplearray) {
-
-    int row;
-
-    pnm_writepaminit(pamP);
-    
-    for (row = 0; row < pamP->height; row++) 
-        pnm_writepamrow(pamP, tuplearray[row]);
-}
-
-
 
 void
 pnm_checkpam(const struct pam *   const pamP, 
@@ -760,6 +784,85 @@ pnm_checkpam(const struct pam *   const pamP,
 
 
 
+void 
+pnm_maketuplergb(const struct pam * const pamP,
+                 tuple              const tuple) {
+
+    if (allocationDepth(pamP) < 3)
+        pm_error("allocation depth %u passed to pnm_maketuplergb().  "
+                 "Must be at least 3.", allocationDepth(pamP));
+
+    if (pamP->depth < 3)
+        tuple[2] = tuple[1] = tuple[0];
+}
+
+
+
+void 
+pnm_makerowrgb(const struct pam * const pamP,
+               tuple *            const tuplerow) {
+    
+    if (pamP->depth < 3) {
+        unsigned int col;
+
+        if (allocationDepth(pamP) < 3)
+            pm_error("allocation depth %u passed to pnm_makerowrgb().  "
+                     "Must be at least 3.", allocationDepth(pamP));
+        
+        if (strncmp(pamP->tuple_type, "RGB", 3) != 0) {
+            for (col = 0; col < pamP->width; ++col) {
+                tuple const thisTuple = tuplerow[col];
+                thisTuple[2] = thisTuple[1] = thisTuple[0];
+            }
+        }
+    }
+}
+
+
+
+void 
+pnm_makearrayrgb(const struct pam * const pamP,
+                 tuple **           const tuples) {
+
+    if (pamP->depth < 3) {
+        unsigned int row;
+        if (allocationDepth(pamP) < 3)
+            pm_error("allocation depth %u passed to pnm_makearrayrgb().  "
+                     "Must be at least 3.", allocationDepth(pamP));
+        
+        for (row = 0; row < pamP->height; ++row) {
+            tuple * const tuplerow = tuples[row];
+            unsigned int col;
+            for (col = 0; col < pamP->width; ++col) {
+                tuple const thisTuple = tuplerow[col];
+                thisTuple[2] = thisTuple[1] = thisTuple[0];
+            }
+        }
+    }
+}
+
+
+
+void
+pnm_getopacity(const struct pam * const pamP,
+               bool *             const haveOpacityP,
+               unsigned int *     const opacityPlaneP) {
+
+    /* Design note; If use of this information proliferates, we should
+       probably add it to struct pam as convenience values analogous to
+       bytes_per_sample.
+    */
+    if (strcmp(pamP->tuple_type, "RGB_ALPHA") == 0) {
+        *haveOpacityP = TRUE;
+        *opacityPlaneP = PAM_TRN_PLANE;
+    } else if (strcmp(pamP->tuple_type, "GRAYSCALE_ALPHA") == 0) {
+        *haveOpacityP = TRUE;
+        *opacityPlaneP = PAM_GRAY_TRN_PLANE;
+    } else
+        *haveOpacityP = FALSE;
+}
+
+
 
 /*=============================================================================
    pm_system() Standard Input feeder and Standard Output accepter functions.   
@@ -799,6 +902,3 @@ pm_accept_to_pamtuples(int    const pipeToSuckFd,
 
     pm_close(inpam.file);
 }
-
-
-
