@@ -15,6 +15,7 @@
 #include "ppm.h"
 #include "libppm.h"
 #include "mallocvar.h"
+#include "ppmcmap.h"
 
 #define HASH_SIZE 20023
 
@@ -84,19 +85,19 @@ ppm_addtocolorhist( colorhist_vector chv,
 
     /* Search colorhist for the color. */
     for ( i = 0; i < *colorsP; ++i )
-    if ( PPM_EQUAL( chv[i].color, *colorP ) ) {
-        /* Found it - move to new slot. */
-        if ( position > i ) {
-        for ( j = i; j < position; ++j )
-            chv[j] = chv[j + 1];
-        } else if ( position < i ) {
-            for ( j = i; j > position; --j )
-                chv[j] = chv[j - 1];
+        if ( PPM_EQUAL( chv[i].color, *colorP ) ) {
+            /* Found it - move to new slot. */
+            if ( position > i ) {
+                for ( j = i; j < position; ++j )
+                    chv[j] = chv[j + 1];
+            } else if ( position < i ) {
+                for ( j = i; j > position; --j )
+                    chv[j] = chv[j - 1];
+            }
+            chv[position].color = *colorP;
+            chv[position].value = value;
+            return;
         }
-        chv[position].color = *colorP;
-        chv[position].value = value;
-        return;
-    }
     if ( *colorsP < maxcolors ) {
         /* Didn't find it, but there's room to add it; so do so. */
         for ( i = *colorsP; i > position; --i )
@@ -109,6 +110,23 @@ ppm_addtocolorhist( colorhist_vector chv,
 
 
 
+colorhash_table
+ppm_alloccolorhash(void)  {
+    colorhash_table cht;
+    int i;
+
+    MALLOCARRAY(cht, HASH_SIZE);
+    if (cht == NULL)
+        pm_error( "out of memory allocating hash table" );
+
+    for (i = 0; i < HASH_SIZE; ++i)
+        cht[i] = NULL;
+
+    return cht;
+}
+
+
+
 static colorhash_table
 computecolorhash(pixel ** const pixels, 
                  const int cols, const int rows, 
@@ -217,28 +235,15 @@ ppm_computecolorhash2(FILE * const ifp,
 
 
 
-colorhash_table
-ppm_alloccolorhash(void)  {
-    colorhash_table cht;
-    int i;
-
-    MALLOCARRAY(cht, HASH_SIZE);
-    if (cht == NULL)
-        pm_error( "out of memory allocating hash table" );
-
-    for (i = 0; i < HASH_SIZE; ++i)
-        cht[i] = NULL;
-
-    return cht;
-}
-
-
-
 int
 ppm_addtocolorhash(colorhash_table const cht, 
                    const pixel *   const colorP, 
                    int             const value) {
+/*----------------------------------------------------------------------------
+   Add color *colorP to the color hash 'cht' with associated value 'value'.
 
+   Assume the color is not already in the hash.
+-----------------------------------------------------------------------------*/
     int retval;
     colorhist_list chl;
 
@@ -258,6 +263,32 @@ ppm_addtocolorhash(colorhash_table const cht,
 }
 
 
+
+void
+ppm_delfromcolorhash(colorhash_table const cht, 
+                     const pixel *   const colorP) {
+/*----------------------------------------------------------------------------
+   Delete the color *colorP from the colorhahs 'cht', if it's there.
+-----------------------------------------------------------------------------*/
+    int hash;
+    colorhist_list * chlP;
+
+    hash = ppm_hashpixel(*colorP);
+    for (chlP = &cht[hash]; *chlP; chlP = &(*chlP)->next) {
+        if (PPM_EQUAL((*chlP)->ch.color, *colorP)) {
+            /* chlP points to a pointer to the hash chain element we want
+               to remove.
+            */
+            colorhist_list const chl = *chlP;
+            *chlP = chl->next;
+            free(chl);
+            return;
+        }
+    }
+}
+
+
+
 static unsigned int
 colorHashSize(colorhash_table const cht) {
 /*----------------------------------------------------------------------------
@@ -351,18 +382,18 @@ ppm_colorhisttocolorhash(colorhist_vector const chv,
 
 
 int
-ppm_lookupcolor( const colorhash_table cht, const pixel * const colorP )
-    {
+ppm_lookupcolor(colorhash_table const cht, 
+                const pixel *   const colorP) {
     int hash;
     colorhist_list chl;
 
-    hash = ppm_hashpixel( *colorP );
-    for ( chl = cht[hash]; chl != (colorhist_list) 0; chl = chl->next )
-    if ( PPM_EQUAL( chl->ch.color, *colorP ) )
-        return chl->ch.value;
+    hash = ppm_hashpixel(*colorP);
+    for (chl = cht[hash]; chl; chl = chl->next)
+        if (PPM_EQUAL(chl->ch.color, *colorP))
+            return chl->ch.value;
 
     return -1;
-    }
+}
 
 
 
@@ -576,10 +607,6 @@ ppm_findclosestcolor(const pixel * const colormap,
                      const pixel * const pP) {
     
     /* Search colormap for closest match.       */
-    /* algorithm taken from ppmquant.c   -IUW   */
-    int const r1 = PPM_GETR(*pP);
-    int const g1 = PPM_GETG(*pP);
-    int const b1 = PPM_GETB(*pP);
 
     int i;
     int ind;
@@ -589,11 +616,7 @@ ppm_findclosestcolor(const pixel * const colormap,
     ind = -1;
 
     for(i = 0; i < ncolors && bestDist > 0; ++i) {
-        int const r2 = PPM_GETR(colormap[i]);
-        int const g2 = PPM_GETG(colormap[i]);
-        int const b2 = PPM_GETB(colormap[i]);
-
-        unsigned int const dist = SQR(r1 - r2) + SQR(g1 - g2) + SQR(b1 - b2);
+        unsigned int const dist = PPM_DISTANCE(*pP, colormap[i]);
         
         if (dist < bestDist ) {
             ind = i;
